/*
 * Copyright (c) 2022 New Vector Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package im.vector.app.features.pin.lockscreen.crypto

import android.os.Build
import android.util.Base64
import androidx.annotation.VisibleForTesting
import im.vector.app.features.pin.PinCodeStore
import im.vector.app.features.pin.lockscreen.crypto.LockScreenCryptoConstants.LEGACY_PIN_CODE_KEY_ALIAS
import org.matrix.android.sdk.api.securestorage.SecretStoringUtils
import org.matrix.android.sdk.api.util.BuildVersionSdkIntProvider
import java.security.Key
import java.security.KeyStore
import javax.crypto.Cipher
import javax.inject.Inject

/**
 * Used to migrate from the old PIN code key ciphers to a more secure ones.
 */
class PinCodeMigrator @Inject constructor(
        private val pinCodeStore: PinCodeStore,
        private val keyStore: KeyStore,
        private val secretStoringUtils: SecretStoringUtils,
        private val buildVersionSdkIntProvider: BuildVersionSdkIntProvider,
) {

    private val legacyKey: Key get() = keyStore.getKey(LEGACY_PIN_CODE_KEY_ALIAS, null)

    /**
     * Migrates from the old ciphers and [LEGACY_PIN_CODE_KEY_ALIAS] to the [newAlias].
     */
    suspend fun migrate(newAlias: String) {
        if (!keyStore.containsAlias(LEGACY_PIN_CODE_KEY_ALIAS)) return

        val pinCode = getDecryptedPinCode() ?: return
        val encryptedBytes = secretStoringUtils.securelyStoreBytes(pinCode.toByteArray(), newAlias)
        val encryptedPinCode = Base64.encodeToString(encryptedBytes, Base64.NO_WRAP)
        pinCodeStore.savePinCode(encryptedPinCode)
        keyStore.deleteEntry(LEGACY_PIN_CODE_KEY_ALIAS)
    }

    fun isMigrationNeeded(): Boolean = keyStore.containsAlias(LEGACY_PIN_CODE_KEY_ALIAS)

    @VisibleForTesting(otherwise = VisibleForTesting.PRIVATE)
    internal suspend fun getDecryptedPinCode(): String? {
        val encryptedPinCode = pinCodeStore.getPinCode() ?: return null
        val cipher = getDecodeCipher()
        val bytes = cipher.doFinal(Base64.decode(encryptedPinCode, Base64.NO_WRAP))
        return String(bytes)
    }

    private fun getDecodeCipher(): Cipher {
        return when (buildVersionSdkIntProvider.get()) {
            Build.VERSION_CODES.LOLLIPOP, Build.VERSION_CODES.LOLLIPOP_MR1 -> getCipherL()
            else -> getCipherM()
        }.also { it.init(Cipher.DECRYPT_MODE, legacyKey) }
    }

    private fun getCipherL(): Cipher {
        // We cannot mock this in tests as it's tied to the actual cryptographic implementation of the OS version
        val provider = if (Build.VERSION.SDK_INT < Build.VERSION_CODES.M) "AndroidOpenSSL" else "AndroidKeyStoreBCWorkaround"
        val transformation = "RSA/ECB/PKCS1Padding"
        return Cipher.getInstance(transformation, provider)
    }

    private fun getCipherM(): Cipher {
        val transformation = "RSA/ECB/OAEPWithSHA-256AndMGF1Padding"
        return Cipher.getInstance(transformation)
    }
}
